import pandas as pd
import requests
from ecopipeline import ConfigManager
from ecopipeline.extract.APIExtractor import APIExtractor
from datetime import datetime, timedelta


class FieldManager(APIExtractor):
    """APIExtractor for the FieldPop / Field Manager API.

    Recursively splits the time range in half whenever the server returns a 500
    'log size too large' response, down to a minimum of 30-minute windows.
    """

    def __init__(self, config: ConfigManager, start_time: datetime = None, end_time: datetime = None,
                 create_csv: bool = True, csv_prefix: str = ""):
        super().__init__(config, start_time, end_time, create_csv, csv_prefix)

    def raw_data_to_df(self, config: ConfigManager, startTime: datetime = None, endTime: datetime = None) -> pd.DataFrame:
        if startTime is None:
            startTime = datetime(2000, 1, 1, 0, 0, 0)
        if endTime is None:
            endTime = datetime.now()

        api_token = config.get_fm_token()
        device_id = config.get_fm_device_id()
        url = (f"https://www.fieldpop.io/rest/method/fieldpop-api/deviceDataLog"
               f"?happn_token={api_token}&deviceID={device_id}"
               f"&startUTCsec={int(startTime.timestamp())}&endUTCsec={int(endTime.timestamp())}")

        try:
            response = requests.get(url)
            if response.status_code == 200:
                df = pd.DataFrame()
                data = response.json()['data']
                for key, value in data.items():
                    for sensor, records in value.items():
                        sensor_rows = [{'time_pt': entry['time'], sensor: entry['value']} for entry in records]
                        df = pd.concat([df, pd.DataFrame(sensor_rows)])
                if not df.empty:
                    df['time_pt'] = pd.to_datetime(df['time_pt'], unit='s')
                    df.set_index('time_pt', inplace=True)
                    df = df.sort_index()
                    df = df.groupby(df.index).mean()
                return df

            if response.status_code == 500:
                json_message = response.json()
                too_large = 'The log size is too large - please try again with a smaller date range.'
                if ('error' in json_message and 'message' in json_message['error']
                        and json_message['error']['message'] == too_large):
                    if endTime - timedelta(minutes=30) < startTime:
                        print(f"Unable to retrieve data for {startTime} - {endTime}")
                        return pd.DataFrame()
                    mid = startTime + (endTime - startTime) / 2
                    df_1 = self.raw_data_to_df(config, startTime, mid)
                    df_2 = self.raw_data_to_df(config, mid, endTime)
                    df = pd.concat([df_1, df_2])
                    df = df.sort_index()
                    df = df.groupby(df.index).mean()
                    return df

            print(f"Failed to make GET request. Status code: {response.status_code}")
            return pd.DataFrame()
        except Exception as e:
            print(f"An error occurred: {e}")
            return pd.DataFrame()



import configparser
import mysql.connector
import mysql.connector.cursor
import sys
import pandas as pd
import os
import math
import mysql.connector.errors as mysqlerrors
from ecopipeline import ConfigManager
from datetime import datetime, timedelta
import numpy as np

class Loader:
    def __init__(self):
        self.data_map = {
            'int64':'float',
            'int32':'float',
            'float64': 'float',
            'M8[ns]':'datetime',
            'datetime64[ns]':'datetime',
            'object':'varchar(25)',
            'bool': 'boolean'
        }

    def load_database(self, config : ConfigManager, dataframe: pd.DataFrame, table_name: str, dbname : str, auto_log_data_loss : bool = False, primary_key : str = 'time_pt'):
        """
        Loads given pandas DataFrame into a MySQL table overwriting any conflicting data. Uses an UPSERT strategy to ensure any gaps in data are filled.
        Note: will not overwrite values with NULL. Must have a new value to overwrite existing values in database

        Parameters
        ----------  
        config : ecopipeline.ConfigManager
            The ConfigManager object that holds configuration data for the pipeline.
        dataframe: pd.DataFrame
            The pandas DataFrame to be written into the mySQL server.
        config_info: dict
            The dictionary containing the configuration information in the data upload. This can be aquired through the get_login_info() function in this package
        primary_key : str
            The name of the primary key in the database to upload to. Default as 'time_pt'
        table_name : str
            overwrites table name from config_info if needed
        dbname : str = None
            name of the database table
        auto_log_data_loss : bool
            if set to True, a data loss event will be reported if no data exits in the dataframe 
            for the last two days from the current date OR if an error occurs
        config_key : str 
            The key in the config.ini file that points to the minute table data for the site. The name of this table is also the site name.
            

        Returns
        ------- 
        bool: 
            A boolean value indicating if the data was successfully written to the database. 
        """
        # Database Connection
        db_connection, cursor = config.connect_db()
        try:

            # Drop empty columns
            dataframe = dataframe.dropna(axis=1, how='all')  
            
            if(len(dataframe.index) <= 0):
                print(f"Attempted to write to {table_name} but dataframe was empty.")
                ret_value = True
            else:

                print(f"Attempting to write data for {dataframe.index[0]} to {dataframe.index[-1]} into {table_name}")
                if auto_log_data_loss and dataframe.index[-1] < datetime.now() - timedelta(days=3):
                    self.report_data_loss(config, config.get_site_name())
                
                # Get string of all column names for sql insert
                sensor_names = primary_key
                sensor_types = ["datetime"]
                for column in dataframe.columns:
                    sensor_names += "," + column    
                    sensor_types.append(self.data_map[dataframe[column].dtype.name])

                # create SQL statement
                insert_str = "INSERT INTO " + table_name + " (" + sensor_names + ") VALUES ("
                for column in dataframe.columns:
                    insert_str += "%s, "
                insert_str += "%s)"
                
                existing_rows_list = []

                # create db table if it does not exist, otherwise add missing columns to existing table
                if not self.check_table_exists(cursor, table_name, dbname):
                    if not self.create_new_table(cursor, table_name, sensor_names.split(",")[1:], sensor_types[1:], primary_key=primary_key): #split on colums and remove first column aka time_pt
                        ret_value = False
                        raise Exception(f"Could not create new table {table_name} in database {dbname}")
                else:
                    try:
                        # find existing times in database for upsert statement
                        cursor.execute(
                            f"SELECT {primary_key} FROM {table_name} WHERE {primary_key} >= '{dataframe.index.min()}'")
                        # Fetch the results into a DataFrame
                        existing_rows = pd.DataFrame(cursor.fetchall(), columns=[primary_key])

                        # Convert the primary_key column to a list
                        existing_rows_list = existing_rows[primary_key].tolist()

                    except mysqlerrors.Error:
                        print(f"Table {table_name} has no data.")

                    missing_cols, missing_types = self.find_missing_columns(cursor, dataframe, dbname, table_name)
                    if len(missing_cols):
                        if not self.create_new_columns(cursor, table_name, missing_cols, missing_types):
                            print("Unable to add new columns due to database error.")
                
                updatedRows = 0
                for index, row in dataframe.iterrows():
                    time_data = row.values.tolist()
                    #remove nans and infinites
                    time_data = [None if (x is None or pd.isna(x)) else x for x in time_data]
                    time_data = [None if (x == float('inf') or x == float('-inf')) else x for x in time_data]

                    if index in existing_rows_list:
                        statement, values = self._generate_mysql_update(row, index, table_name, primary_key)
                        if statement != "":
                            cursor.execute(statement, values)
                            updatedRows += 1
                    else:
                        cursor.execute(insert_str, (index, *time_data))

                db_connection.commit()
                print(f"Successfully wrote {len(dataframe.index)} rows to table {table_name} in database {dbname}. {updatedRows} existing rows were overwritten.")
                ret_value = True
        except Exception as e:
            print(f"Unable to load data into database. Exception: {e}")
            if auto_log_data_loss:
                self.report_data_loss(config, config.get_site_name())
            ret_value = False

        db_connection.close()
        cursor.close()
        return ret_value
    
    def report_data_loss(self, config : ConfigManager, site_name : str = None):
        """
        Logs data loss event in event database (assumes one exists) as a DATA_LOSS_COP event to 
        note that COP calculations have been effected

        Parameters
        ----------  
        config : ecopipeline.ConfigManager
            The ConfigManager object that holds configuration data for the pipeline.
        site_name : str
            the name of the site to correspond the events with. If left blank will default to minute table name

        Returns
        ------- 
        bool: 
            A boolean value indicating if the data was successfully written to the database. 
        """
        # Drop empty columns

        dbname = config.get_db_name()
        table_name = "site_events"
        if site_name is None:
            site_name = config.get_site_name()
        error_string = "Error processing data. Please check logs to resolve."

        print(f"logging DATA_LOSS_COP into {table_name}")

        # create SQL statement
        insert_str = "INSERT INTO " + table_name + " (start_time_pt, site_name, event_detail, event_type, summary_filtered, optim_filtered, last_modified_date, last_modified_by) VALUES "
        insert_str += f"('{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}','{site_name}','{error_string}','DATA_LOSS_COP', true, true, '{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}','automatic_upload')"

        existing_rows = pd.DataFrame({
            'id' : []
        })

        connection, cursor = config.connect_db() 

        # create db table if it does not exist, otherwise add missing columns to existing table
        if not self.check_table_exists(cursor, table_name, dbname):
            print(f"Cannot log data loss. {table_name} does not exist in database {dbname}")
            return False
        else:
            try:
                # find existing times in database for upsert statement
                cursor.execute(
                    f"SELECT id FROM {table_name} WHERE end_time_pt IS NULL AND site_name = '{site_name}' AND event_type = 'DATA_LOSS_COP'")
                # Fetch the results into a DataFrame
                existing_rows = pd.DataFrame(cursor.fetchall(), columns=['id'])

            except mysqlerrors.Error as e:
                print(f"Retrieving data from {table_name} caused exception: {e}")
        try:
            
            if existing_rows.empty:
                cursor.execute(insert_str)
                connection.commit()
                print("Successfully logged data loss.")
            else:
                print("Data loss already logged.")
        except Exception as e:
            # Print the exception message
            print(f"Caught an exception when uploading to site_events table: {e}")
        connection.close()
        cursor.close()
        return True

    def check_table_exists(self, cursor : mysql.connector.cursor.MySQLCursor, table_name: str, dbname: str) -> int:
        """
        Check if the given table name already exists in database.

        Parameters
        ---------- 
        cursor : mysql.connector.cursor.MySQLCursor
            Database cursor object and the table name.
        table_name : str 
            Name of the table
        dbname : str
            Name of the database

        Returns
        ------- 
        int: 
            The number of tables in the database with the given table name.
            This can directly be used as a boolean!
        """

        cursor.execute(f"SELECT count(*) "
                    f"FROM information_schema.TABLES "
                    f"WHERE (TABLE_SCHEMA = '{dbname}') AND (TABLE_NAME = '{table_name}')")

        num_tables = cursor.fetchall()[0][0]
        return num_tables
    
    def create_new_table(self, cursor : mysql.connector.cursor.MySQLCursor, table_name: str, table_column_names: list, table_column_types: list, primary_key: str = "time_pt", has_primary_key : bool = True) -> bool:
        """
        Creates a new table in the mySQL database.

        Parameters
        ---------- 
        cursor : mysql.connector.cursor.MySQLCursor
            A cursor object and the name of the table to be created.
        table_name : str
            Name of the table
        table_column_names : list
            list of columns names in the table must be passed.
        primary_key: str
            The name of the primary index of the table. Should be a datetime. If has_primary_key is set to False, this will just be a column not a key.
        has_primary_key : bool
            Set to False if the table should not establish a primary key. Defaults to True

        Returns
        ------- 
        bool: 
            A boolean value indicating if a table was sucessfully created. 
        """
        if(len(table_column_names) != len(table_column_types)):
            raise Exception("Cannot create table. Type list and Field Name list are different lengths.")

        create_table_statement = f"CREATE TABLE {table_name} (\n{primary_key} datetime,\n"

        for i in range(len(table_column_names)):
            create_table_statement += f"{table_column_names[i]} {table_column_types[i]} DEFAULT NULL,\n"
        if has_primary_key:
            create_table_statement += f"PRIMARY KEY ({primary_key})\n"

        create_table_statement += ");"
        cursor.execute(create_table_statement)

        return True
    
    def find_missing_columns(self, cursor : mysql.connector.cursor.MySQLCursor, dataframe: pd.DataFrame, dbname: str, table_name: str):
        """
        Finds the column names which are not in the database table currently but are present
        in the pandas DataFrame to be written to the database. If communication with database
        is not possible, an empty list will be returned meaning no column will be added. 

        Parameters
        ---------- 
        cursor : mysql.connector.cursor.MySQLCursor 
            A cursor object and the name of the table to be created.
        dataframe : pd.DataFrame
            the pandas DataFrame to be written into the mySQL server. 
        dbname : str
            Name of the database
        data_type : str
            the header name corresponding to the table you wish to write data to.  

        Returns
        ------- 
        list: 
            list of column names which must be added to the database table for the pandas 
            DataFrame to be properly written into the database. 
        """

        try:
            cursor.execute(f"SELECT column_name FROM information_schema.columns WHERE table_schema = '"
                                f"{dbname}' AND table_name = '"
                                f"{table_name}'")
        except mysqlerrors.DatabaseError as e:
            print("Check if the mysql table to be written to exists.", e)
            return [], []
        
        current_table_names = list(cursor.fetchall())
        current_table_names = [name[0] for name in current_table_names]
        df_names = list(dataframe.columns)
        
        cols_to_add = [sensor_name for sensor_name in df_names if sensor_name not in current_table_names]
        data_types = [dataframe[column].dtype.name for column in cols_to_add]
        
        data_types = [self.data_map[data_type] for data_type in data_types]
        
        return cols_to_add, data_types


    def create_new_columns(self, cursor : mysql.connector.cursor.MySQLCursor, table_name: str, new_columns: list, data_types: str):
        """
        Create the new, necessary column in the database. Catches error if communication with mysql database
        is not possible.

        Parameters
        ----------  
        cursor : mysql.connector.cursor.MySQLCursor
            A cursor object and the name of the table to be created.
        config_info : dict
            The dictionary containing the configuration information.
        data_type : str
            the header name corresponding to the table you wish to write data to.  
        new_columns : list
            list of columns that must be added to the database table.

        Returns
        ------- 
        bool:
            boolean indicating if the the column were successfully added to the database. 
        """
        alter_table_statements = [f"ALTER TABLE {table_name} ADD COLUMN {column} {data_type} DEFAULT NULL;" for column, data_type in zip(new_columns, data_types)]

        for sql_statement in alter_table_statements:
            try:
                cursor.execute(sql_statement)
            except mysqlerrors.DatabaseError as e:
                print(f"Error communicating with the mysql database: {e}")
                return False

        return True
    
    def _generate_mysql_update(self, row, index, table_name, primary_key):
        statement = f"UPDATE {table_name} SET "
        statment_elems = []
        values = []
        for column, value in row.items():
            if not value is None and not pd.isna(value) and not (value == float('inf') or value == float('-inf')):
                statment_elems.append(f"{column} = %s")
                values.append(value)

        if values:
            statement += ", ".join(statment_elems)
            statement += f" WHERE {primary_key} = '{index}';"
        else:
            statement = ""

        return statement, values


import pandas as pd
from ecopipeline import ConfigManager
from datetime import datetime
from ecopipeline.extract.FileProcessor import FileProcessor


class MSACSVProcessor(FileProcessor):
    """FileProcessor for MSA CSV files.

    MSA files use a DateEpoch(secs) column for the timestamp, which is converted
    from UTC to the configured time zone. When mb_prefix is True, column names are
    prefixed with the filename stem.
    """

    def __init__(self, config: ConfigManager, start_time: datetime = None, end_time: datetime = None,
                 filename_date_format: str = "%Y%m%d%H%M%S", file_prefix: str = "", data_sub_dir: str = "",
                 date_string_start_idx: int = -17, date_string_end_idx: int = -3,
                 mb_prefix: bool = False, time_zone: str = 'US/Pacific'):
        self.mb_prefix = mb_prefix
        self.time_zone = time_zone
        super().__init__(config, ".csv", start_time, end_time,
                         filename_date_format=filename_date_format, file_prefix=file_prefix,
                         data_sub_dir=data_sub_dir, date_string_start_idx=date_string_start_idx,
                         date_string_end_idx=date_string_end_idx, round_time_index=True)

    def _read_file_into_df(self, file_name: str) -> pd.DataFrame:
        data = pd.read_csv(file_name)
        if len(data) != 0:
            data['time_pt'] = pd.to_datetime(data['DateEpoch(secs)'], unit='s', utc=True)
            data['time_pt'] = data['time_pt'].dt.tz_convert(self.time_zone).dt.tz_localize(None)
            data.set_index('time_pt', inplace=True)
            data.drop(columns='DateEpoch(secs)', inplace=True)
            if self.mb_prefix:
                prefix = file_name.split('.')[0].split('/')[-1]
                data = data.rename(columns={col: f"{prefix}{col}".replace(' ', '_').replace('*', '_') for col in data.columns})
        return data



import pandas as pd
import numpy as np
import datetime as datetime
from ecopipeline import ConfigManager
import re
import mysql.connector.errors as mysqlerrors
from datetime import datetime, timedelta
from ecopipeline.extract.FileProcessor import FileProcessor

class CSVProcessor(FileProcessor):
    def __init__(self, config: ConfigManager, start_time: datetime = None, end_time: datetime = None, raw_time_column: str = 'DateTime',
                 time_column_format: str = '%Y/%m/%d %H:%M:%S', filename_date_format: str = "%Y%m%d%H%M%S", file_prefix: str = "", data_sub_dir: str = "",
                 date_string_start_idx: int = -17, date_string_end_idx: int = -3):
        super().__init__(config, ".csv", start_time, end_time, raw_time_column, time_column_format, filename_date_format,
                         file_prefix, data_sub_dir, date_string_start_idx, date_string_end_idx)
        
    def _read_file_into_df(self, file_name : str) -> pd.DataFrame:
        data = super()._read_file_into_df(file_name)
        if len(data) != 0:
            data['time_pt'] = pd.to_datetime(data[self.raw_time_column], format=self.time_column_format)
            data.set_index('time_pt', inplace=True)
                
        return data


import pandas as pd
import numpy as np
import datetime as datetime
from ecopipeline import ConfigManager
import re
import mysql.connector.errors as mysqlerrors
from datetime import datetime, timedelta
import os

class FileProcessor:
    def __init__(self, config : ConfigManager, extension: str, start_time: datetime = None, end_time: datetime = None, raw_time_column : str = 'DateTime',
                 time_column_format : str ='%Y/%m/%d %H:%M:%S', filename_date_format : str = "%Y%m%d%H%M%S", file_prefix : str = "", data_sub_dir : str = "",
                 date_string_start_idx : int = -17, date_string_end_idx : int = -3, round_time_index : bool = False):
        """
        Parameters
        ----------  
        config : ecopipeline.ConfigManager
            The ConfigManager object that holds configuration data for the pipeline 
        extension : str
            File extension of raw data files as string (e.g. ".csv", ".gz", ...)
        start_time: datetime
            The point in time for which we want to start the data extraction from. This 
            is local time from the data's index.
        end_time: datetime
            The point in time for which we want to end the data extraction. This 
            is local time from the data's index.
        data_sub_dir : str
            defaults to an empty string. If the files being accessed are in a sub directory of the configured data directory, use this parameter to point there.
            e.g. if the data files you want to extract are in "path/to/data/DENT/" and your configured data directory is "path/to/data/", put "DENT/" as the data_sub_dir
        file_prefix : str
            File name prefix of raw data file if only file names with a certain prefix should be processed.
        
        Returns
        ------- 
        list[str]: 
            list of filenames 
        """
        self.extension = extension
        self.start_time = start_time
        self.end_time = end_time
        self.raw_time_column = raw_time_column
        self.time_column_format = time_column_format
        self.filename_date_format = filename_date_format
        self.date_string_start_idx = date_string_start_idx
        self.date_string_end_idx = date_string_end_idx
        self.file_prefix = file_prefix
        self.data_sub_dir = data_sub_dir
        self.round_time_index = round_time_index
        self.raw_df = pd.DataFrame()
        try:
            filenames = self.extract_files(config)
            self.raw_df = self.raw_files_to_df(filenames)
        except Exception as e:
            print(f"File extraction failed: {e}")
            raise e
        
    def get_raw_data(self) -> pd.DataFrame:
        return self.raw_df

    def extract_files(self, config: ConfigManager) -> list[str]:
        """
        Function returns a list of all file names in the processors assigned directory with file names that fall between the start and end date (if such dates exist)
        
        Returns
        ------- 
        list[str]: 
            list of filenames 
        """
        os.chdir(os.getcwd())
        filenames = []
        full_data_path = f"{config.data_directory}{self.data_sub_dir}"
        for file in os.listdir(full_data_path):
            if file.endswith(self.extension) and file.startswith(self.file_prefix):
                full_filename = os.path.join(full_data_path, file)
                filenames.append(full_filename)

        if not self.start_time is None:
            filenames = self.extract_new(filenames)

        return filenames
    
    def extract_new(self, filenames: list[str]) -> list[str]:
        
        endTime_int = self.end_time
        startTime_int = int(self.start_time.strftime(self.filename_date_format))
        if not self.end_time is None:
            endTime_int = int(self.end_time.strftime(self.filename_date_format)
                            )
        return_list = list(filter(lambda filename: int(filename[self.date_string_start_idx:self.date_string_end_idx]) >= startTime_int and (endTime_int is None or int(filename[self.date_string_start_idx:self.date_string_end_idx]) < endTime_int), filenames))
        return return_list
    
    def _read_file_into_df(self, file_name : str) -> pd.DataFrame:
        data = pd.read_csv(file_name)
        return data

    def raw_files_to_df(self, filenames : list[str]) -> pd.DataFrame:
        temp_dfs = []
        for file in filenames:
            try:
                data = self._read_file_into_df(file)
                if len(data) != 0:
                    temp_dfs.append(data)
            except FileNotFoundError:
                print("File Not Found: ", file)
                continue
            except Exception as e:
                print(f"Error reading {file}: {e}")
                continue
                    
        if len(temp_dfs) <= 0:
            print("No data for timefarme.")
            return pd.DataFrame()
        
        df = pd.concat(temp_dfs, ignore_index=False)

        if self.round_time_index:
            df.index = df.index.floor('min')
            df = df.groupby(df.index).mean(numeric_only=True)
            df.sort_index(inplace=True)

        return df


import pandas as pd
from ecopipeline import ConfigManager
from datetime import datetime, timedelta
import re
from ecopipeline.extract.FileProcessor import FileProcessor


class ModbusCSVProcessor(FileProcessor):
    """FileProcessor for Modbus CSV files (e.g. Acquisuite).

    Reads standard CSVs whose first column is a time(UTC) timestamp. Column names
    are prefixed with the filename stem so data from multiple devices can be merged
    without column collisions. Index is floored to the minute after concat.
    """

    def __init__(self, config: ConfigManager, start_time: datetime = None, end_time: datetime = None,
                 raw_time_column: str = 'time(UTC)', filename_date_format: str = "%Y%m%d%H%M%S",
                 file_prefix: str = "", data_sub_dir: str = "",
                 date_string_start_idx: int = -17, date_string_end_idx: int = -3):
        super().__init__(config, ".csv", start_time, end_time,
                         raw_time_column=raw_time_column,
                         filename_date_format=filename_date_format, file_prefix=file_prefix,
                         data_sub_dir=data_sub_dir, date_string_start_idx=date_string_start_idx,
                         date_string_end_idx=date_string_end_idx, round_time_index=True)
        
    def extract_new(self, filenames: list[str]) -> list[str]:
        
        base_date = datetime(1970, 1, 1)
        file_dates = [pd.Timestamp(base_date + timedelta(seconds = int(re.search(r'\.(.*?)_', filename.split("/")[-1]).group(1), 16))) for filename in filenames] #convert decihex to dates, these are in utc
        
        file_dates_local = [file_date.tz_localize('UTC').tz_localize(None) for file_date in file_dates] #convert utc 
        # if timeZone == None:
        #     file_dates_local = [file_date.tz_localize('UTC').tz_localize(None) for file_date in file_dates] #convert utc 
        # else:
        #     file_dates_local = [file_date.tz_localize('UTC').tz_convert(timezone(timeZone)).tz_localize(None) for file_date in file_dates] #convert utc to local zone with no awareness

        return_list = [filename for filename, local_time in zip(filenames, file_dates_local) if local_time > self.start_time and (self.end_time is None or local_time < self.end_time)]
        return return_list

    def _read_file_into_df(self, file_name: str) -> pd.DataFrame:
        data = pd.read_csv(file_name)
        if len(data) != 0:
            if self.raw_time_column in data.columns:
                prefix = file_name.split('/')[-1].split('.')[0]
                data[self.raw_time_column] = pd.to_datetime(data[self.raw_time_column])
                data = data.set_index(self.raw_time_column)
                data = data.rename(columns={col: f"{prefix}_{col}".replace(' ', '_') for col in data.columns})
            else:
                print(f"Error reading {file_name}: No '{self.raw_time_column}' column found.")
                return pd.DataFrame()
        return data



from .CSVProcessor import CSVProcessor
from .JSONProcessor import JSONProcessor
from .ModbusCSVProcessor import ModbusCSVProcessor
from .DentCSVProcessor import DentCSVProcessor
from .FlowCSVProcessor import FlowCSVProcessor
from .MSACSVProcessor import MSACSVProcessor
from .EGaugeCSVProcessor import EGaugeCSVProcessor
from .SmallPlanetCSVProcessor import SmallPlanetCSVProcessor



import pandas as pd
import gzip
import json
import requests
from ecopipeline import ConfigManager
from ecopipeline.extract.APIExtractor import APIExtractor
from datetime import datetime, timedelta


class Skycentrics(APIExtractor):
    """APIExtractor for the Skycentrics API.

    Pulls data day-by-day between start_time and end_time, normalises the JSON
    sensor records into a pivot table, and rounds 59:59 timestamps up to the
    next minute.
    """

    def __init__(self, config: ConfigManager, start_time: datetime = None, end_time: datetime = None,
                 create_csv: bool = True, csv_prefix: str = "", time_zone: str = 'US/Pacific'):
        self.time_zone = time_zone
        super().__init__(config, start_time, end_time, create_csv, csv_prefix)

    def raw_data_to_df(self, config: ConfigManager, startTime: datetime = None, endTime: datetime = None) -> pd.DataFrame:
        if endTime is None:
            endTime = datetime.utcnow()
        if startTime is None:
            startTime = endTime - timedelta(days=1)

        temp_dfs = []
        time_parser = startTime
        while time_parser < endTime:
            time_parse_end = time_parser + timedelta(days=1)
            start_time_str = time_parser.strftime('%Y-%m-%dT%H:%M:%S')
            end_time_str = time_parse_end.strftime('%Y-%m-%dT%H:%M:%S')
            skycentrics_token, date_str = config.get_skycentrics_token(
                request_str=f'GET /api/devices/{config.api_device_id}/data?b={start_time_str}&e={end_time_str}&g=1 HTTP/1.1',
                date_str=None)
            response = requests.get(
                f'https://api.skycentrics.com/api/devices/{config.api_device_id}/data?b={start_time_str}&e={end_time_str}&g=1',
                headers={'Date': date_str, 'x-sc-api-token': skycentrics_token, 'Accept': 'application/gzip'})
            if response.status_code == 200:
                decompressed_data = gzip.decompress(response.content)
                json_data = json.loads(decompressed_data)
                norm_data = pd.json_normalize(json_data, record_path=['sensors'], meta=['time'], meta_prefix='response_')
                if len(norm_data) != 0:
                    norm_data['time_pt'] = pd.to_datetime(norm_data['response_time'], utc=True)
                    norm_data['time_pt'] = norm_data['time_pt'].dt.tz_convert(self.time_zone)
                    norm_data = pd.pivot_table(norm_data, index='time_pt', columns='id', values='data')
                    for i in range(len(norm_data.index)):
                        if norm_data.index[i].minute == 59 and norm_data.index[i].second == 59:
                            norm_data.index.values[i] = norm_data.index[i] + pd.Timedelta(seconds=1)
                    temp_dfs.append(norm_data)
            else:
                print(f"Failed to make GET request. Status code: {response.status_code} {response.json()}")
            time_parser = time_parse_end

        if len(temp_dfs) > 0:
            return pd.concat(temp_dfs, ignore_index=False)
        print("No Skycentrics data retrieved for time frame.")
        return pd.DataFrame()
